'''# pyncm
本模块提供`Get/Set/DumpCurrentSession`,`LoadNewSessionFromDump` 以管理请求
API 使用请参见 `pyncm.apis` 文档 
'''

from .utils.crypto import Crypto
from typing import Text, Union
from time import time
import requests,logging,json

__version__ = "1.6.3.6"

class Session(requests.Session):   
    '''Represents an API session'''
    HOST          = "https://music.163.com/"
    UA_DEFAULT    = 'Mozilla/5.0 (linux@github.com/greats3an/pyncm) Chrome/But.It.Was.Actually.PyNCM.%s' % __version__ # XXX this is just stupid
    UA_EAPI       = 'NeteaseMusic/7.2.24.1597753235(7002024);Dalvik/2.1.0 (Linux; U; Android 11; Pixel 2 XL Build/RP1A.200720.009)'
    UA_LINUX_API  = 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.90 Safari/537.36'    
    CONFIG_EAPI   = {
        # template EAPI config pulled from version 7.2.24, which would be overriden if some of the keys were specified in cookies 
        'appver': '7.2.24','buildver':'7002024','channel':'offical','deviceId': Crypto.RandomString(8),
        'mobilename' : 'Pixel2XL','os': 'android','osver':'10.1','resolution': '2712x1440','versioncode': '240'         
    }

    def __init__(self,*a,**k):        
        super().__init__(*a,**k)
        self.headers = {
            "Content-Type"  : "application/x-www-form-urlencoded",  
            "User-Agent"    : self.UA_DEFAULT,
            "Referer"       : self.HOST
        }
        self.login_info = {'success': False,'tick': time(), 'content': None}
        self.csrf_token = ''

    def request(self, method: str, url: Union[str, bytes, Text], *a,**k) -> requests.Response:
        '''Initiates & fires a request

        Args:
            method (str): HTTP Verb
            url (Union[str, bytes, Text]): Full / partial API URL

        Returns:
            requests.Response: A requests.Response object
        '''    
        if url[:4] != 'http':url = self.HOST + url                
        return super().request(method,url,*a,**k)  

    # region auth IO
    _AUTH_SYMBOLS = {
        # symbols for loading/reloading authentication info
        'login_info':(lambda self:getattr(self,'login_info'),lambda self,v:setattr(self,'login_info',v)),
        'csrf_token':(lambda self:getattr(self,'csrf_token'),lambda self,v:setattr(self,'csrf_token',v)),
        'cookies':   (
            lambda self:
                [{'name': c.name, 'value': c.value, 'domain': c.domain, 'path': c.path} for c in getattr(self,'cookies')],
            lambda self,cookies:
                [getattr(self,'cookies').set(**cookie) for cookie in cookies]
        )
    }
    def dump(self) -> dict:
        '''Dumps current authencation info as dictionary'''
        return {name:self._AUTH_SYMBOLS[name][0](self) for name in self._AUTH_SYMBOLS.keys()}
    def load(self,dumped):
        '''Loads current authencation info from dictionary'''
        for k,v in dumped.items():self._AUTH_SYMBOLS[k][1](self,v)
        return True
    # endregion 

class SessionManager():
    '''Manages session switching / loading / dumping'''
    def __init__(self) -> None:
        self.session = Session()
    def get(self):
        return self.session
    def set(self,session):
        self.session = session
    # region Session serialization
    @staticmethod
    def stringify(session : Session) -> str:
        return Crypto.EapiCrypto('pyncm',json.dumps(session.dump()))['params']
    @staticmethod
    def parse(dump : str) -> Session:
        session = Session() # a new session WILL be created
        dump = Crypto.HexCompose(dump)
        dump = Crypto.EapiDecrypt(dump).decode()
        dump = dump.split('-36cd479b6b5-')
        assert dump[0] == 'pyncm' # check magic
        session.load(json.loads(dump[1])) # loading config dict in
        return session
    # endregion
       
logger = logging.getLogger('ncm')
'''Logger to be used by local modules'''
sessionManager              = SessionManager()
'''Manages active session'''
def GetCurrentSession() -> Session:
    '''Retrives current active session'''
    return sessionManager.get()
def SetCurrentSession(session : Session):
    '''Sets current active session'''
    sessionManager.set(session)
def SetNewSession():
    '''Creates and sets new session'''
    sessionManager.set(Session())
def LoadSessionFromString(dump : str) -> Session:
    '''Loads a session from dumped string'''
    session = SessionManager.parse(dump)
    return session
def DumpSessionAsString(session : Session) -> str:
    '''Dumps session as encrypted string'''
    return SessionManager.stringify(session)
from .apis import *


'''Implementation of some of NE's crypto functions'''
import base64,json,math,ctypes
from Crypto.Cipher import AES
from Crypto.Random import random
from hashlib import md5
from hashlib import md5
from time import time

class RSAPublicKey():
    def __init__(self, n, e):
        '''RSA pubkey pubkey & modulus pair,values are interpeted as integers'''
        self.n = int(n, 16)
        self.e = int(e, 16)
# region Constant values
WEAPI_AES_KEY    = "0CoJUm6Qyw8W8jud" # cbc
WEAPI_AES_IV     = "0102030405060708" # cbc
WEAPI_RSA_PUBKEY = RSAPublicKey(
    "00e0b509f6259df8642dbc35662901477df22677ec152b5ff68ace615bb7b725152b3ab17a876aea8a5aa76d2e417629ec4ee341f56135fccf695280104e0312ecbda92557c93870114af6c9d05c4f7f0c3685b7a46bee255932575cce10b424d813cfe4875d3e82047b97ddef52741d546b8e289dc6935b3ece0462db0a22b8e7",
    "10001" # textbook rsa without padding
)
LINUXAPI_AES_KEY = "rFgB&h#%2?^eDg:Q" # ecb
EAPI_DIGEST_SALT = "nobody%(url)suse%(text)smd5forencrypt"
EAPI_DATA_SALT   = "%(url)s-36cd479b6b5-%(text)s-36cd479b6b5-%(digest)s"
EAPI_AES_KEY     = "e82ckenh8dichen8" # ecb

BASE62           = 'PJArHa0dpwhvMNYqKnTbitWfEmosQ9527ZBx46IXUgOzD81VuSFyckLRljG3eC'
BASE64 = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'

_obfs_Rb = 'LPc4uQNptC2A6y.R90DOBIroS+qnx/eb3FM8fW1UZG7VmwvksgjhaTKzlXdiYEHJ'
_obfs_db = 5

WM_DID = 'Eyj1ZVEWep5ERQEFFVMuN0xNxLvKykXT' 
'''DeviceID - for checkToken (watchman.js)'''
# endregion

# region checkToken (watchman.js) stuff
# region Bitwise operations
def _int(v):return ctypes.c_int(v).value
def _lshift(v,bs):
    return _int(v << (bs & 0x1F))
def _rshift(v,bs):
    return _int(v >> (bs & 0x1F))
# Bit shifting only takes care of LS 5 bits
def _truncate(v):
    return v & 0xFFFFFFFF    
# MDN : All javascript bit operations are done on 32 bit integers
def _xor(v1,v2): return _int( _truncate(v1) ^ _truncate(v2))
# endregion
# region MD5
def _hash_Yb(e):
    '''MD5 hasher'''
    from . import Crypto
    return Crypto.HashDigest(e)
def _hex_Zb(e):
    '''MD5 digest hexi-fier'''
    from . import Crypto
    return Crypto.HexDigest(e)
# endregion
# region Special functions
def _map_int_C(e):
    '''Integer mapping - (-inf,inf) -> [-127,127]'''
    if e < -128:
        return _map_int_C(128 - (-128 - e))
    if (e >= -128 and e <= 127):
        return e
    if (e > 127):
        return _map_int_C(-129 + e - 127)
def _to_bytes_X(b):
    '''Integer to bytes - int(32 bits) -> [4 bytes]'''
    return [_map_int_C(b >> 24 & 255),_map_int_C(b >> 16 & 255),_map_int_C(b >> 8 & 255),_map_int_C(b & 255)]
def _encodeURIComponent_Z(e):
    '''encodeURIComponent()'''
    return e
def _unescape_ac(e):
    '''unescape()'''
    return e
def _charcodes_ya(e):    
    '''String to charcode array'''
    return [ord(char) for char in _encodeURIComponent_Z(e)] if isinstance(e,str) else e
def _to_hex_Ub(e):
    '''Char to hexdecimal [00,FF] string'''
    c = "0123456789abcdef"
    return c[e >> 4 & 15] + c[e & 15]
def _str_to_hex_gb(e):
    '''String to hexstring converter'''
    return ''.join([_to_hex_Ub(a) for a in e])
def _xor_with_array_La(e):    
    '''Simple XOR cipherer'''
    d = [31, 125, -12, 60, 32, 48]
    e = _charcodes_ya(e)
    g = []
    for q in range(0,len(e)):         
        g.append(_map_int_C(_xor(e[q],d[q % len(d)])))
        g[q] = _map_int_C(-g[q])
    return _str_to_hex_gb(g)
def _shuffle_ga(e, c, d, r, g):
    '''Array shuffler'''
    for h in range(0,g):d[r + h] = e[c + h]
def _hexstr_to_int_hb(e):
    '''Hexstring to integer array'''
    c = []
    for r in range(0,len(e),2):
        h = _lshift(int(e[r], 16),4)
        m = int(e[r + 1], 16)
        c.append(_map_int_C(h + m))
    return c    
def _obfuscate_array_Qb(e, c, d=0, r=_obfs_Rb, g=_obfs_db):
    '''Integer array to obfuscated string'''
    h, m,g = 0,[],str(g)
    def push(*a):m.extend(a)
    if d == 1:
        d = e[c];
        h = 0;
        push(r[ d >> 2 & 63], r[(d << 4 & 48) + (h >> 4 & 15)], g, g)
    elif d == 2:
        d = e[c];
        h = e[c + 1];
        e = 0;
        push(r[d >> 2 & 63], r[(d << 4 & 48) + (h >> 4 & 15)], r[(h << 2 & 60) + (e >> 6 & 3)], g)
    elif d == 3:
        d = e[c];
        h = e[c + 1];
        e = e[c + 2];
        push(r[d >> 2 & 63], r[(d << 4 & 48) + (h >> 4 & 15)], r[(h << 2 & 60) + (e >> 6 & 3)], r[e & 63])
    else:
        return None
    return ''.join(m)
def _double_mapping_Sb(e, c=[], d=_obfs_db):
    '''Characters mapped to another array'''
    r = 3
    h = 0
    g = []
    while (h < len(e)):
        if (h + r <= len(e)):
            g.append(_obfuscate_array_Qb(e, h, r, c, d))
            h += r
        else:
            g.append(_obfuscate_array_Qb(e, h, len(e) - h, c, d));
            break
    return "".join(g)
def _map_string_Tb(a):
    return _double_mapping_Sb(a,BASE64,'=')
def _generate_OTP_b():
    '''OTP tokenizer'''
    e = int(time() * 1000) # it seems that this timestamp should within 1mo of today's time
    c = math.floor(e / 4294967296)
    d = e % 4294967296
    e = _to_bytes_X(c)
    d = _to_bytes_X(d)
    c = [ 0 ] * 8    
    _shuffle_ga(e, 0, c, 0, 4);
    _shuffle_ga(d, 0, c, 4, 4);  
    d = [ 0 ] * 8 # should be random nums of [0-255] * 8
    e = [ 0 ] * 16
    for g in range(0,len(c) * 2):
        if (g % 2 == 0):
            f = int(g / 2)
            e[g] = e[g] | (d[f] & 16) >> 4 | (d[f] & 32) >> 3 | (d[f] & 64) >> 2 | (d[f] & 128) >> 1 | (c[f] & 16) >> 3 | (c[f] & 32) >> 2 | (c[f] & 64) >> 1 | (c[f] & 128) >> 0
        else:
            f = math.floor(g / 2)
            e[g] = e[g] | (d[f] & 1) << 0 | (d[f] & 2) << 1 | (d[f] & 4) << 2 | (d[f] & 8) << 3 | (c[f] & 1) << 1 | (c[f] & 2) << 2 | (c[f] & 4) << 3 | (c[f] & 8) << 4
        e[g] = _map_int_C(e[g])
    c = _str_to_hex_gb(e)
    c = _hex_Zb(_hash_Yb(_unescape_ac(_encodeURIComponent_Z(c + "dAWsBhCqtOaNLLJ25hBzWbqWXwiK99Wd"))))
    c = _hexstr_to_int_hb(c[:16])
    c = _map_string_Tb(c+e)
    return c
def _generate_config_chiper_bc(c,f=1,g=WM_DID):
    '''Generates cipher of our local config'''
    t = _unescape_ac(_encodeURIComponent_Z(str(f) + str(g) + str(c) + "WoeTpXnDDPhiAvsJUUIY3RdAo2PKaVwi"))
    t = _hash_Yb(t)
    return _xor_with_array_La(json.dumps({
        'r': f,      # config.va - should be 1 all the time
        'd': g,      # g || "",device ID
        'b': c,
        't': _map_string_Tb(t)
    }))
# endregion    
# endregion    

class Crypto():
    '''The cryptography ...things'''
    # region Base cryptograhpy methods
    # region Utility
    @staticmethod
    def RandomString(len, chars=BASE62):
        '''Generates random string of `len` chars within a selected number of chars'''
        return ''.join([random.choice(chars) for i in range(0, len)])
    @staticmethod
    def HexDigest(data : bytearray):
        '''Digests a `bytearray` to a hex string'''
        return ''.join([hex(d)[2:].zfill(2) for d in data])
    @staticmethod
    def HexCompose(hexstr : str):
        '''Composes a hex string back to a `bytearray`'''
        return bytearray([int(hexstr[i:i+2],16) for i in range(0,len(hexstr),2)])
    @staticmethod
    def HashDigest(text):
        '''Digests 128 bit md5 hash'''
        HASH = md5(text.encode('utf-8'))
        return HASH.digest()        
    @staticmethod
    def HashHexDigest(text):
        '''Digests 128 bit md5 hash,then digest it as a hexstring'''
        return Crypto.HexDigest( Crypto.HashDigest(text) )
    # endregion
    # region Cryptos    
    @staticmethod
    def checkToken(deviceId=WM_DID):
        '''Generates `checkToken` parameter
        
        Args:
            deviceId (str, optional): XXX: Device ID fetched. Defaults to WM_DID.

        Returns:
            str
        '''
        return _generate_config_chiper_bc(_generate_OTP_b(),g=deviceId)
    @staticmethod
    def AESEncrypt(
        data:str, key:str, mode=AES.MODE_ECB , iv=''       
    ):
        '''Basic AES encipher function'''
        def pad(data,blocksize=AES.block_size):return data + (blocksize - len(data) % blocksize) * chr(blocksize - len(data) % blocksize)        
        if mode in [AES.MODE_EAX,AES.MODE_ECB]:
            encryptor = AES.new(key.encode(), mode)
        else:    
            encryptor = AES.new(key.encode(), mode, iv.encode())
        encrypt_aes = encryptor.encrypt(str.encode(pad(data)))
        return encrypt_aes
    @staticmethod
    def AESDecrypt(
        data:str,key:str,mode=AES.MODE_ECB,iv=''
    ):
        '''Basic AES decipher funtion'''
        def unpad(data):
            pad = data[-1]
            if not pad in range(0,AES.block_size):return data # data isn't padded
            return data[:-pad]
        if mode in [AES.MODE_EAX,AES.MODE_ECB]:
            decryptor = AES.new(key.encode(), mode)
        else:    
            decryptor = AES.new(key.encode(), mode, iv.encode())
        decrypted_aes = decryptor.decrypt(data)
        return unpad(decrypted_aes)
    @staticmethod
    def RSAEncrypt(data:str, pubkey: RSAPublicKey, reverse=True):
        '''Signle-block textbook RSA encrpytion (c ≡ n ^ e % N),encodes text to hexstring first'''
        n = data if not reverse else reversed(data)        
        n, e, N = int(''.join(n).encode('utf-8').hex(), 16), pubkey.e, pubkey.n
        r = pow(n,e,N) # n ** e % N - modular exponetiation
        return Crypto.HexCompose(hex(r)[2:].zfill(256))
    # endregion
    # endregion
    # region Netease crypto
    '''source:core.js thingy'''
    @staticmethod
    def WeapiCrypto(params, aes_key2=None):
        '''Used in web & PC client'''
        aes_key2 = aes_key2 or Crypto.RandomString(16)
        params = str(params)
        # 1st go,encrypt the text with aes_key and aes_iv
        params = str(base64.encodebytes(Crypto.AESEncrypt(
            data=params, key=WEAPI_AES_KEY, iv=WEAPI_AES_IV,mode=AES.MODE_CBC)), encoding='utf-8')
        # 2nd go,encrypt the ENCRYPTED text again,with the 2nd key and aes_iv
        params = str(base64.encodebytes(Crypto.AESEncrypt(
            data=params, key=aes_key2, iv=WEAPI_AES_IV,mode=AES.MODE_CBC)), encoding='utf-8')
        # 3rd go,generate RSA encrypted encSecKey
        encSecKey = Crypto.HexDigest(Crypto.RSAEncrypt(aes_key2, WEAPI_RSA_PUBKEY))        
        return {
            'params': params,
            'encSecKey': encSecKey
        }
    '''source:
        - decompilation of `libpoison.so` : https://juejin.im/post/6844903586879520775
        - Binaryify/NeteaseCloudMusicApi  : https://github.com/Binaryify/NeteaseCloudMusicApi/blob/master/util/crypto.js'''
    @staticmethod
    def LinuxCrypto(params):
        '''Used in desktop linux clients'''
        params = str(params)
        # Single-pass AES-128 ECB crypto
        return {
            'eparams':Crypto.HexDigest(Crypto.AESEncrypt(params,key=LINUXAPI_AES_KEY,mode=AES.MODE_ECB))
        }
    @staticmethod
    def EapiCrypto(url,params):
        '''Used in mobile and PC clients'''
        url,params = str(url),str(params)
        digest = Crypto.HashHexDigest(EAPI_DIGEST_SALT % {'url':url,'text':params})
        params = EAPI_DATA_SALT % ({'url':url,'text':params,'digest':digest})
        return {
            'params':Crypto.HexDigest(Crypto.AESEncrypt(params,key=EAPI_AES_KEY,mode=AES.MODE_ECB))
        }
    @staticmethod
    def EapiDecrypt(cipher):
        '''Used in mobile clients'''
        cipher = bytearray(cipher) if isinstance(cipher,str) else cipher
        return Crypto.AESDecrypt(cipher,EAPI_AES_KEY,mode=AES.MODE_ECB) if cipher else cipher
    # endregion


import logging,time
truncate_length = 64
logger = logging.getLogger('NCMHelper')

def TrackHelperProperty(default=None):
    def preWrapper(func):
        @property
        def wrapper(*a,**k):
            try:
                return func(*a,**k)
            except:
                logger.warn('Error while getting track attribute %s,using fallback value %s' % (func.__name__,default))
                return default
        return wrapper
    return preWrapper

class TrackHelper():
    '''Helper class for handling generic track objects'''

    def __init__(self, track_dict) -> None:
        self.track = track_dict

    @TrackHelperProperty()
    def ID(self):
        '''Track ID'''
        return self.track['id']

    @TrackHelperProperty()
    def TrackPublishTime(self):
        '''The publish year of the track'''
        epoch = self.track['publishTime'] / 1000 # stored as unix timestamp though only the year was ever useful
        return time.gmtime(epoch).tm_year

    @TrackHelperProperty()
    def TrackNumber(self):
        '''The # of the track'''
        return self.track['no']

    @TrackHelperProperty(default='Unknown')
    def TrackName(self):
        '''The name of the track'''
        assert self.track['name'] != None
        return self.track['name']

    @TrackHelperProperty(default='Unknown')
    def AlbumName(self):
        '''The name of the album'''
        if self.track['al']['id']:
            return self.track['al']['name']
        else:
            return self.track['pc']['alb']

    @TrackHelperProperty(default='https://p1.music.126.net/UeTuwE7pvjBpypWLudqukA==/3132508627578625.jpg')
    def AlbumCover(self):
        '''The cover URL of the track's album'''
        al = self.track['al'] if 'al' in self.track.keys() else self.track['album']
        if al['id']:
            return al['picUrl']
        else:        
            return 'https://music.163.com/api/img/blur/' + self.track['pc']['cid'] 
            # source:PC version's core.js

    @TrackHelperProperty(default=['Various Artists'])
    def Artists(self):
        '''All the artists' names as a list'''
        ar = self.track['ar'] if 'ar' in self.track.keys() else self.track['artists']
        ret = [_ar['name'] for _ar in ar ]
        if not ret.count(None):
            return ret
        else:
            return [self.track['pc']['ar']] # for NCM cloud-drive stored audio

    @TrackHelperProperty()
    def Title(self):
        '''A formatted title for this song'''
        return f'{self.TrackName} - {",".join(self.Artists)}'
    _illegal_chars = set('\x00\\/:*?"<>|')
    @TrackHelperProperty()
    def SanitizedTitle(self):
        '''Sanitized title for FS compatibility; Limits max length to 200 chars,and substitutes illegal chars with their full-width counterparts'''
        def _T(s,l=100):
            return ''.join([c if not c in self._illegal_chars else chr(ord(c)+0xFEE0) for c in s])[:l]
        return f'{_T(self.TrackName,100)} - {_T(",".join(self.Artists),100)}'


'''`__main__` entry for PyNCM'''
from pyncm import GetCurrentSession,__version__,logger
from pyncm.utils.lrcparser import LrcParser
from pyncm.utils.helper import TrackHelper
from pyncm.apis import login,track,playlist,album
import tqdm

from queue import Queue
from concurrent.futures import ThreadPoolExecutor
from threading import Thread
from time import sleep
from os.path import join,exists
from os import remove,makedirs

from logging import getLogger,basicConfig
import sys,argparse,base64,urllib.parse

__desc__ = '''PyNCM 网易云音乐下载工具'''
max_workers = 4
    
BITRATES = {'standard':96000,'high':320000,'lossless':3200000}
# Key-Value classes
class BaseKeyValueClass:
    def __init__(self,**kw) -> None:
        for k,v in kw.items():self.__setattr__(k,v)

class BaseDownloadTask(BaseKeyValueClass):
    id : int
    url : str
    dest : str

class LyricsDownloadTask(BaseDownloadTask):
    id : int
    dest : str
    lrc_blacklist : set

class TrackDownloadTask(BaseKeyValueClass):
    song : TrackHelper    
    cover : BaseDownloadTask
    lyrics : BaseDownloadTask
    resource : BaseDownloadTask
    resource_info : dict
    lyrics_exclude : set
    save_as : str

class TaskPoolExecutorThread(Thread):
    @staticmethod
    def tag_audio(tHelper : TrackHelper,audio : str,cover : str):
        from mutagen.flac import FLAC, Picture
        from mutagen.id3 import ID3, APIC
        from mutagen.mp3 import EasyMP3
        from mutagen import easymp4
        from mutagen.mp4 import MP4, MP4Cover
        from mutagen.oggvorbis import OggVorbis    
        def write_keys(song):
            # Write trackdatas
            song['title'] = tHelper.TrackName
            song['artist'] = tHelper.Artists
            song['album'] = tHelper.AlbumName
            song['tracknumber'] = str(tHelper.TrackNumber)
            song['date'] = str(tHelper.TrackPublishTime)
            song.save()    
        format = audio.split('.')[-1].lower()
        if format == 'm4a':
            # Process m4a files: Apple’s MP4 (aka M4A, M4B, M4P)
            song = easymp4.EasyMP4(audio)
            write_keys(song)
            if exists(cover):
                song = MP4(audio)
                # Write cover image
                song['covr'] = [MP4Cover(open(cover, 'rb').read())]
                song.save()
        elif format == 'mp3':
            # Process mp3 files: MPEG audio stream information and tags
            song = EasyMP3(audio)
            write_keys(song)
            if exists(cover):
                song = ID3(audio)
                # Write cover image
                song.add(APIC(encoding=3, mime='image/jpeg', type=3,
                                desc='Cover', data=open(cover, 'rb').read()))
                song.save()
        elif format == 'flac':
            # Process FLAC files:Free Lossless Audio Codec
            song = FLAC(audio)
            write_keys(song)
            if exists(cover):
                pic = Picture()
                pic.data = open(cover, 'rb').read()
                pic.mime = 'image/jpeg'
                song.add_picture(pic)
                song.save()
        elif format == 'ogg':
            # Process OGG files:Ogg Encapsulation Format
            song = OggVorbis(audio)
            write_keys(song)
            if exists(cover):
                pic = Picture()
                pic.data = open(cover, 'rb').read()
                pic.mime = 'image/jpeg'
                song["metadata_block_picture"] = [base64.b64encode(pic.write()).decode('ascii')]
                song.save()
        return True

    def download_by_url(self,url,dest,xfer=False):
        # Downloads generic content
        response = GetCurrentSession().get(url,stream=True)                        
        length = int(response.headers.get('content-length'))
        
        with open(dest,'wb') as f:
            for chunk in response.iter_content(1024 * 2 ** 10):
                self.xfered += len(chunk)
                if xfer:
                    self.finished_tasks += len(chunk) / length # task [0,1]
                f.write(chunk) # write every 1MB read            
        return dest

    def __init__(self,*a,**k):
        super().__init__(*a,**k)
        self.finished_tasks : float = 0
        self.xfered = 0        
        self.task_queue = Queue()

    def run(self):
        def execute(task : TrackDownloadTask):
            try:
                if not exists(task.resource.dest):makedirs(task.resource.dest)
                # Downloding source audio
                dest_src = self.download_by_url(task.resource.url,join(task.resource.dest,task.save_as+'.'+task.resource_info['type']),xfer=True)
                # Downloading cover
                dest_cvr = self.download_by_url(task.cover.url,join(task.cover.dest,'%s.jpg' % task.cover.id))
                # Downloading & Parsing lyrics
                dest_lrc = join(task.lyrics.dest,task.save_as + '.lrc')
                lrc    = LrcParser()
                dLyrics = track.GetTrackLyrics(task.lyrics.id)
                for k in set(dLyrics.keys()) & ({'lrc','tlyric','romalrc'} - task.lyrics.lrc_blacklist): # Filtering LRCs
                    lrc.LoadLrc(dLyrics[k]['lyric'])    
                lrc_text = lrc.DumpLyrics()                       
                if lrc_text:open(dest_lrc,'w',encoding='utf-8').write(lrc_text)
                # Tagging the audio
                self.tag_audio(task.song,dest_src,dest_cvr)
                # Cleaning up
                remove(dest_cvr)               
            except Exception as e:
                logger.warning('下载未完成 %s - %s' % (task.song.Title,e))
        with ThreadPoolExecutor(max_workers=max_workers) as executor:        
            while True:
                task = self.task_queue.get()            
                future = executor.submit(execute,task)
                future.add_done_callback(lambda future:self.task_queue.task_done())
# Subroutines
class Subroutine:
    '''Generic subroutine
    
    Subroutines are `callable`,upon called with `ids`,one
    queues tasks with all given arguments via `put_func` callback
    '''
    def __init__(self,args,put_func) -> None:
        self.args = args
        self.put = put_func

def create_subroutine(sub_type) -> Subroutine:
    '''Dynamically creates subroutine callable by string specified'''
    class Playlist(Subroutine):
        def forIds(self,ids):
            dDetails = track.GetTrackDetail(ids)['songs']
            dAudios   = track.GetTrackAudio(ids,bitrate=BITRATES[self.args.quality])['data']
            
            dDetails = sorted(dDetails,key=lambda song:song['id'])
            dAudios = sorted(dAudios,key=lambda song:song['id'])               
            queued = 0
            for dDetail,dAudio in zip(dDetails,dAudios):
                song = TrackHelper(dDetail)
                logger.info('单曲 #%d / %d - %s - %s (%dkbps)' % (queued+1,len(dDetails),song.Title,song.AlbumName,dAudio['br'] // 1000))        
                tSong = TrackDownloadTask(
                    song = song,
                    cover = BaseDownloadTask(id=song.ID,url=song.AlbumCover,dest=self.args.output),                    
                    resource = BaseDownloadTask(id=song.ID,url=dAudio['url'],dest=self.args.output),
                    lyrics = LyricsDownloadTask(id=song.ID,dest=self.args.output,lrc_blacklist=set(self.args.lyric_no)),
                    resource_info = dAudio,
                    save_as = self.args.template.format(**{
                        'id':song.ID,
                        'year':song.TrackPublishTime,
                        'no': song.TrackNumber,
                        'track' : song.TrackName,
                        'album' : song.AlbumName,
                        'title' : song.SanitizedTitle,
                        'artists' : ' / '.join(song.Artists),
                    })
                )          
                self.put(tSong)
                queued += 1
            return queued
        def __call__(self,ids): 
            queued = 0           
            for _id in ids:
                dList = playlist.GetPlaylistInfo(_id)
                logger.info('歌单 ：%s' % dList['playlist']['name'])
                queued += self.forIds([tid['id'] for tid in dList['playlist']['trackIds']])
            return queued

    class Album(Playlist):
        def __call__(self,ids):    
            queued = 0    
            for _id in ids:    
                dList = album.GetAlbumInfo(_id)
                logger.info('专辑 ：%s' % dList['album']['name'])
                queued += self.forIds([tid['id'] for tid in dList['songs']])
            return queued

    class Song(Playlist):
        def __call__(self,ids):            
            return self.forIds(ids)

    return {
        'song':Song,'playlist':Playlist,'album':Album
    }[sub_type]

def parse_sharelink(url):
    split = urllib.parse.urlsplit(url)
    dest  = split.path.split('/')[-1]
    query = urllib.parse.parse_qs(split.query)
    return dest,query

def parse_args():    
    '''Setting up __main__ argparser'''
    parser = argparse.ArgumentParser(description=__desc__, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('url',metavar='链接',help='网易云音乐分享链接')
    group = parser.add_argument_group('下载')
    group.add_argument('--template',metavar='模板',help=r'''保存文件名模板
    参数：    
        id     - 网易云音乐资源 ID
        year   - 出版年份
        no     - 专辑中编号
        album  - 专辑标题
        track  - 单曲标题        
        title  - 完整标题
        artists- 艺术家名
    例：
        {track} - {artists} 等效于 {title}''',default=r'{title}')
    group.add_argument('--quality',metavar='音质',choices=list(BITRATES.keys()),help=r'''音频音质（高音质需要 CVIP）
    参数：
        lossless - “无损”
        high     - 较高
        standard - 标准''',default='standard')
    group.add_argument('--output',metavar='输出',default='.',help='输出文件夹')
    group = parser.add_argument_group('歌词')
    group.add_argument('--lyric-no',metavar='跳过歌词',help=r'''跳过某些歌词类型的合并
    参数：
        lrc    - 源语言歌词
        tlyric - 翻译后歌词
        romalrc- 罗马音歌词
    例：
        --lyric-no tlyric --lyric-no romalrc 将只下载源语言歌词''', choices=['lrc','tlyric','romalrc'], default='',nargs='+')
    group = parser.add_argument_group('登陆')
    group.add_argument('--phone',metavar='手机',default='',help='网易账户手机号')
    group.add_argument('--pwd',metavar='密码',default='',help='网易账户密码')
    args = parser.parse_args()        
    dest,query = parse_sharelink(args.url)

    ids = query['id']
    dest = dest
                
    return ids,args,dest    

def __main__():
    ids,args,dest  = parse_args()
    if args.phone and args.pwd:        
        login.LoginViaCellphone(args.phone,args.pwd)
        logger.info('账号 ：%s (VIP %s)' % (
            GetCurrentSession().login_info['content']['profile']['nickname'],
            GetCurrentSession().login_info['content']['profile']['vipType'])
        )
    executor = TaskPoolExecutorThread()
    executor.setDaemon(True)
    executor.start()   
    
    def enqueue_task(task):        
        executor.task_queue.put(task)

    subroutine = create_subroutine(dest)(args,enqueue_task)
    total_queued = subroutine(ids) # Enqueue tasks

    _tqdm = tqdm.tqdm(bar_format='已完成 {desc}: {percentage:.1f}%|{bar}| {n:.2f}/{total_fmt} [{elapsed}<{remaining}')
    _tqdm.total = total_queued
    def report():                        
        _tqdm.desc = _tqdm.format_sizeof(executor.xfered,suffix='B',divisor=1024)        
        _tqdm.update(min(executor.finished_tasks,total_queued) - _tqdm.n)
        return True    

    while executor.task_queue.unfinished_tasks >= 0:
        report() and sleep(.5)
        if executor.task_queue.unfinished_tasks == 0:break

    return 0

if __name__ == '__main__':
    logger = getLogger()
    basicConfig(level='INFO',format='[%(levelname).4s] %(message)s')    
    sys.exit(__main__())


'''Another lrc file parser'''
import re
from collections import defaultdict

def LrcProperty(tagname):    
    def wrapper(func):
        @property
        def _wrapper(self):
            if not hasattr(self,tagname):return Exception(tagname)
            try:
                return getattr(self,tagname)
            except:
                # So this is how we enmurate all the attributes on startup
                # As no values are assigned (not inintalized),they will surely cause errors
                # Making it fallback to another value.Here,we fallback to its raw tagname
                return Exception(tagname)
        @_wrapper.setter
        def _wrapper(self,v):
            setattr(self,tagname,v)
        return _wrapper  
    return wrapper

class LrcRegexes:
    LIDTag_= re.compile(r'(?<=\[)[^\[\]]*(?=\])')
    LIDTag_Type = re.compile(r'[a-z]{2,}(?=:)')
    LIDTag_Content = re.compile(r'(?<=[a-z]{2}:).*')
    LLyrics_ = re.compile(r'[^\[\]]*$')
    LBrackets = re.compile(r'(?<=\[).*(?=\])')
    LTimestamp = re.compile(r'\d*[\.,:]\d*[\.,:]\d*')
# region Static methods
# Timestamp parsers
def stamp2tag(timestamp):
    mm = int(timestamp / 60)
    ss = int((timestamp - mm * 60))
    xx = int((timestamp - mm * 60 - ss) * 100) # We'd use standard 100th of a second here
    mm,ss,xx = str(mm).rjust(2,'0'),str(ss).rjust(2,'0'),str(xx).rjust(2,'0')
    return f'{mm}:{ss}.{xx}'                    
    
def tag2stamp(IDTag):
    IDTag = ''.join(LrcRegexes.LTimestamp.findall(IDTag))
    if not IDTag:return None
    div = IDTag.split(':') 
    if (len(div)==2):
        mm,ss = div
        ss,xx = ss.split('.') # xx is hunderth of a second,but NE didn't think so
    else:
        mm,ss,xx = div
    timestamp = int(mm) * 60 + int(ss) + int(xx) * (0.1 ** len(xx)) # <- workaround
    return timestamp
# endregion

class LrcParser:
    '''Parses lrc into mutable dict-like objects'''
    # region Properties from Wikipedia [https://en.wikipedia.org/wiki/LRC_(file_format)]
    @LrcProperty('ar')
    def Artist(self):pass
    @LrcProperty('al')
    def Album(self):pass
    @LrcProperty('ti')
    def Title(self):pass
    @LrcProperty('au')
    def Author(self):pass
    @LrcProperty('length')
    def Length(self):pass
    @LrcProperty('by')
    def LRCAuthor(self):pass
    @LrcProperty('offset')
    def Offset(self):pass
    @LrcProperty('re')
    def Program(self):pass
    @LrcProperty('ve')
    def ProgramVersion(self):pass
    # endregion
    
    def __init__(self,lrc=''):
        '''Takes lyrics in `LRC` format,then provides lyrics based on timestamps'''
        # Parsing lrc,line by line
        def EnmurateAttributes():
            for m in dir(self):
                if any(f in m for f in ['__','Add','Load','Clear','Find','Update','Dump','lyrics']):continue
                yield (m,str(getattr(self,m)))        
        self.lrcAttributes = list(EnmurateAttributes())
        # This function will only work when no attributes are defined
        self.lyrics = defaultdict(list)
        if not lrc:
            # empty input,we are creating lyrics then
            return
        else:self.LoadLrc(lrc)

    @property
    def lyrics_sorted(self):
        '''Returns sorted version of the lyrics'''
        return defaultdict(list,sorted(self.lyrics.items()))        

    def LoadLrc(self,lrc):
        '''Loads a LRC formmated lryics file'''
        for line in lrc.split('\n'):
            IDTag           =         LrcRegexes.LIDTag_       .findall(line)
            IDTagType       = ''.join(LrcRegexes.LIDTag_Type   .findall(''.join(IDTag)))
            IDTagContent    = ''.join(LrcRegexes.LIDTag_Content.findall(''.join(IDTag)))
            Lyrics          = ''.join(LrcRegexes.LLyrics_      .findall(line))
            if IDTagType:
                # Tag's type is set,write as class attribute
                setattr(self,IDTagType,IDTagContent)
            elif IDTag:
                # Tag's type is not set but we got the values,treat as lyrics
                # We'll use the timestamp (into seconds) as lyrics' keys,as hashtables can handle that            
                try:
                    for _IDTag in IDTag:
                        # Some LRC lyrics would pile a bunch of timestamps in on line,like 
                        #   [00:01.12][00:08.12]Yeah                        
                        timestamp = tag2stamp(_IDTag)
                        if timestamp is not None and timestamp >= 0:
                            if not isinstance(self.Offset,Exception):timestamp += float(self.Offset)
                            if Lyrics:self.lyrics[timestamp].append((_IDTag,Lyrics)) # Ignore empty lines
                except:
                    pass
        self.lyrics = self.lyrics_sorted # sort stuff once loaded

    def AddLyrics(self,timestamp,value):
        # Add 1 or multiple line(s) of lyrics with the same timestamp in seconds 
        if not isinstance(value,list):value = [value]
        for v in value:self.lyrics[timestamp].append((stamp2tag(timestamp),v))
        self.lyrics = self.lyrics_sorted # this isn't too much for previously sorted items though
    
    def ClearLyrics(self):
        # Clears the lyrics buffer
        self.lyrics.clear()

    def UpdateLyrics(self,iterable,timestamp_function,lyrics_function):
        '''This function takes an iterable,a timestamp function,and a lyrics function
        
        And for every item in the iterable,timestamp and lyrics will be fetch via the said functions,then get added to our buffer
        '''
        for line in iterable:
            timestamp = timestamp_function(line)
            lyrics    = lyrics_function(line)
            self.AddLyrics(timestamp,lyrics)

    def DumpLyrics(self,delimiter='\t'):
        '''Format current lyrics buffer then spits out a LRC formatted string'''
        lrc = ''
        # Adding tags
        for prop,attr in self.lrcAttributes:
            value = getattr(self,prop)
            if not isinstance(value,Exception): # If such value do exist
                lrc += f'[{attr}:{value}]' + '\n'        
        # Adding lyrics
        for timestamp,lyrics in self.lyrics.items(): # write the sorted one
            Lyrics = []
            for tag,_lyric in lyrics:
                Lyrics.append(_lyric)
            lrc += '\n' + f'[{stamp2tag(timestamp)}]{delimiter.join(Lyrics)}' 
        return lrc # Done
    
    @staticmethod
    def Find(lyrics,timestamp):
        '''Finds closest match in our hashable
            
            lyrics  :   From `lrcparser.lyrics_sorted`

            Returns `(timestamp_seconds,lyrics,indexof)`

            Returns None if the match isn't inside the window
        '''
        def search(val,src:list,l,r):    
            '''Binary serach'''            
            pivot = (l+r)>>1
            if (l >= r or pivot == l or pivot == r):return pivot            
            in_between = (pivot<=0 or (src[pivot-1] <= val)) and val <= src[pivot]            
            if in_between:            
                if src[pivot] == val:
                    return pivot
                elif src[pivot - 1] == val:
                    return pivot - 1
                else:
                    return pivot
            return search(val,src,pivot,r) if src[pivot] < val else search(val,src,l,pivot)        
        timestamps = list(lyrics.keys())
        index      = search(timestamp,timestamps,0,len(lyrics))        
        return (timestamps[index],lyrics[timestamps[index]],index)